//   Copyright (c)  2022  John Abbott and Anna M. Bigatti
//   Original author:  Nico Mexis  (2022)

//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/SparsePolyOps-Cyclotomicity.H"
#include "CoCoA/BigInt.H"
#include "CoCoA/BigIntOps.H"
#include "CoCoA/error.H"
#include "CoCoA/factor.H"
#include "CoCoA/interrupt.H"
#include "CoCoA/MachineInt.H"
#include "CoCoA/NumTheory-factor.H"
#include "CoCoA/NumTheory-misc.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/ring.H"
#include "CoCoA/RingQQ.H"
#include "CoCoA/RingZZ.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/SparsePolyOps-RingElem.H"
#include "CoCoA/verbose.H"

namespace CoCoA
{

    /**
     * Provides a simple-to-compute bound for the inverse
     * of the Euler Totient function. All numbers k with
     * phi(k) = n will satisfy k < InversePhiBound(n).
     * See https://oeis.org/A355667
     */
    BigInt InversePhiBound(const unsigned long n)
    {
        if (n < 8) return 3 * BigInt(n);
        if (n < 48) return 4 * BigInt(n);
        if (n < 5760) return 5 * BigInt(n);
        if (n < 36495360L) return 6 * BigInt(n);
        if (n < 1854081073152000L) return 7 * BigInt(n);
        // higher bound as fallback for BigInt
        return 3 * FloorSqrt(power(n, 3));
        // could also be extended using OEIS sequence if needed
    }

    /**
     * Ported directly from C5 code without optimizations.
     * Used for testing current implementation of IsCyclotomic
     * and IsCyclotomic2 functions.
     */
    RingElem gcdList(const std::vector<RingElem> &L)
    {
        const size_t n = L.size();
        if (n == 0)
            CoCoA_THROW_ERROR(ERR::Empty, "gcdList");

        RingElem g = L[0];
        for (size_t i = 1; i < n; ++i)
        {
            CheckForInterrupt("gcdList");
            g = gcd(g, L[i]);
            // @JA: In coclib.cpkg5, IsOne could maybe be included?
            if (IsOne(g))
                break;
        }
        return g;
    }

    /**
     * Ported directly from C5 code without optimizations.
     * Used for testing current implementation of IsCyclotomic
     * and IsCyclotomic2 functions. Checks are missing!
     */
    RingElem cyclotomic(const ring &P, const long K)
    {
        const RingElem &x = indet(P, 0);
        if (K == 1)
            return x - 1; // K is 1

        std::vector<long> primes = factor(K).myFactors();
        const size_t primesSize = primes.size();
        if (primesSize == 1)
            // K is prime or prime power
            return (power(x, K) - 1) / (power(x, K / primes[0]) - 1);

        //  L := divisors(K);
        //  If L = [1] Then Return sum([x^I| I In 0..(K-1)]); EndIf;// K is prime
        //  M := MaxDiv(L);
        //  If len(M) = 1 Then Return (x^K-1)/(x^M[1]-1); EndIf;    // K is prime power

        std::vector<RingElem> temp(primesSize);
        for (size_t i = 0; i < primesSize; ++i)
        {
            CheckForInterrupt("cyclotomic");
            temp[i] = (power(x, K) - 1) / (power(x, K / primes[i]) - 1);
        }

        return gcdList(temp); // general case
    }

    unsigned long IsCyclotomicGraeffe(ConstRefRingElem f);

    namespace // anonymous
    {
        /**
         * Checks if f has a square root and then if that
         * square root is cyclotomic. Returns true iff
         * both conditions are met.
         * For private use. Hence, no arg checking.
         */
        bool HasCyclotomicSqrt(ConstRefRingElem f)
        {
            VerboseLog VERBOSE("HasCyclotomicSqrt");
            factorization<RingElem> fac = SqFreeFactor(f);
            VERBOSE(80) << "fac = " << fac << std::endl;
            if (!IsOne(fac.myRemainingFactor()))
                return false;
            RingElem sqrt = one(owner(f));

            const std::vector<RingElem> &factors = fac.myFactors();
            const std::vector<long> &mults = fac.myMultiplicities();
            for (size_t i = 0; i < factors.size(); ++i)
            {
                CheckForInterrupt("HasCyclotomicSqrt");

                const long mult = mults[i];

                if (IsOdd(mult))
                    return false;

                sqrt *= power(factors[i], mult / 2);
            }
            VERBOSE(50) << "sqrt = " << sqrt << std::endl;

            return IsCyclotomicGraeffe(sqrt);
        }
    }

    /**
     * Graeffe approach
     * NOT UPDATED AND NOT SUPPORTED!
     */
    unsigned long IsCyclotomicGraeffe(ConstRefRingElem f)
    {
        VerboseLog VERBOSE("IsCyclotomicGraeffe");
        const ring &Px = owner(f);
        if (!IsSparsePolyRing(Px))
            CoCoA_THROW_ERROR(ERR::NotSparsePolyRing, "IsCyclotomicGraeffe");
        const ring &P = CoeffRing(Px);
        const long IndetIndex = UnivariateIndetIndex(f);
        if (IndetIndex < 0)
            CoCoA_THROW_ERROR(ERR::NotUnivariate, "IsCyclotomicGraeffe");
        if (!IsZZ(P) && !IsQQ(P))
            CoCoA_THROW_ERROR(ERR::BadRing, "IsCyclotomicGraeffe");
        const RingElem &x = indet(Px, IndetIndex);

        // Cyclotomic polynomials have LC = 1 and CC = +-1
        if (!IsOne(LC(f)))
            return 0;
        if (!IsOne(abs(ConstantCoeff(f))))
            return 0;

        const long degf = deg(f);
        // deg = 1 -> x +- 1 is cyclotomic
        if (degf == 1)
            return ConstantCoeff(f) > 0 ? 2 : 1;
        // Constant polynomials and ones with odd degree
        // cannot be cyclotomic (except deg = 1; see above!)
        if (degf < 1 || IsOdd(degf))
            return 0;
        // Cyclotomic polynomials are always irreducible
        if (!IsIrred(f))
            return 0;

        const RingElem f1 = graeffe(f);
        VERBOSE(80) << "graeffe = " << f1 << std::endl;

        if (f1 == f)
            return 1;

        const RingElem f2 = PolyAlgebraHom(Px, Px, std::vector<RingElem>(NumIndets(Px), -x))(f);
        VERBOSE(80) << "f2 = " << f2 << std::endl;
        if (f1 == f2 && IsCyclotomicGraeffe(f2))
            return 1;
        VERBOSE(40) << "SqrtTest" << std::endl;
        if (HasCyclotomicSqrt(f1))
            return 1;

        return 0;
    }

    namespace { // anonymous
        RingElem LM(ConstRefRingElem f) {
            return monomial(owner(f), LC(f), LPP(f));
        }
        RingElem CoeffOfSecondTerm(ConstRefRingElem f)
        {
            if (IsConstant(f))
                CoCoA_THROW_ERROR(ERR::BadArg, "CoeffOfSecondTerm");
            RingElem g = f - LM(f);
            if (IsZero(g) || deg(g) != (deg(f)-1))
                return zero(owner(f));
            return LC(g);
        }
    }

    /**
     * Inverse Phi method
     */
    unsigned long IsCyclotomicIter(ConstRefRingElem f)
    {
        VerboseLog VERBOSE("IsCyclotomicIter");
        const ring &Px = owner(f);
        if (!IsSparsePolyRing(Px))
            CoCoA_THROW_ERROR(ERR::NotSparsePolyRing, "IsCyclotomicIter");
        const ring &P = CoeffRing(Px);
        const long IndetIndex = UnivariateIndetIndex(f);
        if (IndetIndex < 0)
            CoCoA_THROW_ERROR(ERR::NotUnivariate, "IsCyclotomicIter");
        if (!IsZZ(P) && !IsQQ(P))
            CoCoA_THROW_ERROR(ERR::BadRing, "IsCyclotomicIter");
        const RingElem &x = indet(Px, IndetIndex);

        // Cyclotomic polynomials have LC = 1 and CC = 1 (except Phi_1)
        if (!IsOne(LC(f))) // TODO: Should we allow LC = -1?
            return 0;
        const long degf = deg(f);
        if (!IsOne(ConstantCoeff(f)))
          return (degf == 1 && IsMinusOne(ConstantCoeff(f)));

        // deg = 1 -> Only possibility is now Phi_2
        if (degf == 1)
            return 2;
        // Constant polynomials and ones with odd degree
        // cannot be cyclotomic (except deg = 1; see above!)
        if (degf < 1 || IsOdd(degf))
            return 0;

        // Second coefficient equal to -Moebius function
        long snd;
        if (!IsConvertible(snd, CoeffOfSecondTerm(f)))
            CoCoA_THROW_ERROR(ERR::ArgTooBig, "IsCyclotomicIter");
        if (uabs(snd) > 1)
            return 0;

        // deg > 1 -> Always palindromic
        if (!IsPalindromic(f))
            return 0;
        // Cyclotomic polynomials are always irreducible
        if (!IsIrred(f))
            return 0;

        BigInt bound = InversePhiBound(degf);
        unsigned long N;
        if (!IsConvertible(N, bound))
            CoCoA_THROW_ERROR(ERR::ArgTooBig, "IsCyclotomicIter");

        VERBOSE(20) << "n bound = " << N << std::endl;
        for (unsigned long n = 3; n <= N; ++n)
        {
            CheckForInterrupt("IsCyclotomicIter");
            VERBOSE(100) << "n = " << n << std::endl;
            if (snd == -MoebiusFn(n) && EulerTotient(n) == degf &&
                IsDivisible(power(x, n) - 1, f))
                return n;
        }

        return 0;
    }

    /**
     * Chooses the right approach for the given polynomial
     */
    unsigned long IsCyclotomic(ConstRefRingElem f)
    {
        // TODO: Are there any criterions?
        return IsCyclotomicIter(f);
    }

}


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SparsePolyOps-Cyclotomicity.C,v 1.3 2022/11/30 15:17:43 abbott Exp $
// $Log: SparsePolyOps-Cyclotomicity.C,v $
// Revision 1.3  2022/11/30 15:17:43  abbott
// Summary: Removed IsCyclotomicGraeffe & IsCyclotomicIter
//
// Revision 1.2  2022/11/24 21:11:29  abbott
// Summary: Added copyright notice
//
//
