//   Copyright (c)  2022  John Abbott and Anna M. Bigatti
//   Original author:  Nico Mexis  (2022)

//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include <set>
#include "CoCoA/SparsePolyOps-LRSDegeneracy.H"
#include "CoCoA/CanonicalHom.H"
#include "CoCoA/factor.H"
#include "CoCoA/factorization.H"
#include "CoCoA/interrupt.H"
#include "CoCoA/NumTheory-gcd.H"
#include "CoCoA/NumTheory-misc.H"
#include "CoCoA/NumTheory-modular.H"
#include "CoCoA/NumTheory-PrimeModSeq.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingQQ.H"
#include "CoCoA/RingZZ.H"
#include "CoCoA/SparsePolyOps-Cyclotomicity.H"
#include "CoCoA/SparsePolyOps-Graeffe.H"
#include "CoCoA/SparsePolyOps-resultant.H"
#include "CoCoA/SparsePolyOps-RingElem.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/verbose.H"

namespace CoCoA
{

  /**
   * Algorithm 1 (naive factor-approach)
   * NOT SUPPORTED AND NOT UPDATED
   */
  unsigned long IsLRSDegenerateNaive(RingElem f)
  {
    VerboseLog VERBOSE("IsLRSDegenerateNaive");
    const ring &Px = owner(f);
    if (!IsSparsePolyRing(Px))
      CoCoA_THROW_ERROR(ERR::NotSparsePolyRing, "IsLRSDegenerateNaive");
    const ring &P = CoeffRing(Px);
    const long IndetIndex = UnivariateIndetIndex(f);
    if (IndetIndex < 0)
      CoCoA_THROW_ERROR(ERR::NotUnivariate, "IsLRSDegenerateNaive");

    if (IsQQ(P))
    {
      const ring &ZZ = RingZZ();
      const SparsePolyRing ZZx = NewPolyRing(ZZ, NewSymbols(NumIndets(Px)));
      const RingHom toZZx = PolyRingHom(Px, ZZx, QQEmbeddingHom(ZZ), indets(ZZx));
      return IsLRSDegenerateNaive(toZZx(f / content(f)));
    }

    if (!IsZZ(P))
      CoCoA_THROW_ERROR(ERR::BadRing, "IsLRSDegenerateNaive");
    const RingElem &x = indet(Px, IndetIndex);

    // Must have non-zero constant coefficient
    f /= gcd(f, power(x, deg(f)));

    const long degf = deg(f);
    if (degf < 2)
      CoCoA_THROW_ERROR("Polynomial does not have two distinct non-zero roots.", "IsLRSDegenerateNaive");

    if (!IsCoprime(f, PolyAlgebraHom(Px, Px, std::vector<RingElem>(NumIndets(Px), -x))(f)))
      return 2;

    const SparsePolyRing Q = NewPolyRing(P, NewSymbols(2));
    const RingElem &xQ = indet(Q, 0);
    const RingElem &yQ = indet(Q, 1);

    // g is a polynomial whose roots are the quotients of the roots of f
    RingElem g = resultant(PolyAlgebraHom(Px, Q, std::vector<RingElem>(NumIndets(Px), xQ * yQ))(f),
                           PolyAlgebraHom(Px, Q, std::vector<RingElem>(NumIndets(Px), yQ))(f), 1);
    g = PolyAlgebraHom(Q, Px, {indet(Px, IndetIndex), one(Px)})(g);
    g /= gcd(g, power(x - 1, degf));

    // TODO: CyclotomicPart could replace this at some point!
    std::vector<RingElem> fac = factor(g).myFactors();
    VERBOSE(50) << "fac = " << fac << std::endl;

    // f is LRS-degenerate if g has a cyclotomic factor
    // other than x-1, which would indicate the quotient
    // of a root and itself
    return any_of(fac.cbegin(), fac.cend(), IsCyclotomic);
  }

  /**
   * Algorithm 2 (Graeffe iteration), used for reducible polynomials
   * NOT SUPPORTED AND NOT UPDATED
   */
  unsigned long IsLRSDegenerateIter(RingElem f)
  {
    VerboseLog VERBOSE("IsLRSDegenerateIter");
    const ring &Px = owner(f);
    if (!IsSparsePolyRing(Px))
      CoCoA_THROW_ERROR(ERR::NotSparsePolyRing, "IsLRSDegenerateIter");
    const ring &P = CoeffRing(Px);
    const long IndetIndex = UnivariateIndetIndex(f);
    if (IndetIndex < 0)
      CoCoA_THROW_ERROR(ERR::NotUnivariate, "IsLRSDegenerateIter");

    if (IsQQ(P))
    {
      const ring &ZZ = RingZZ();
      const SparsePolyRing ZZx = NewPolyRing(ZZ, NewSymbols(NumIndets(Px)));
      const RingHom toZZx = PolyRingHom(Px, ZZx, QQEmbeddingHom(ZZ), indets(ZZx));
      return IsLRSDegenerateIter(toZZx(f / content(f)));
    }

    if (!IsZZ(P))
      CoCoA_THROW_ERROR(ERR::BadRing, "IsLRSDegenerateIter");
    const RingElem &x = indet(Px, IndetIndex);

    // Must have non-zero constant coefficient
    f /= gcd(f, power(x, deg(f)));

    const long degf = deg(f);
    if (degf < 2)
      CoCoA_THROW_ERROR("Polynomial does not have two distinct non-zero roots.", "IsLRSDegenerateIter");

    if (!IsSqFree(f))
      CoCoA_THROW_ERROR("Polynomial is not square-free!", "IsLRSDegenerateIter");

    if (!IsCoprime(f, PolyAlgebraHom(Px, Px, std::vector<RingElem>(NumIndets(Px), -x))(f)))
      return 2;

    BigInt bound = InversePhiBound(degf);
    unsigned long N;
    if (!IsConvertible(N, bound))
      CoCoA_THROW_ERROR(ERR::ArgTooBig, "IsLRSDegenerateIter");

    // Skip first two Graeffe iterations
    GraeffeSeq graeffeSeq(f, N);
    ++graeffeSeq;
    ++graeffeSeq;

    VERBOSE(20) << "n bound = " << N << std::endl;
    for (unsigned long n = 3; n <= N; ++n)
    {
      VERBOSE(100) << "n = " << n << std::endl;

      ++graeffeSeq;
      // Equal to IsSqFree, but without arg checks
      // TODO: Is IsSqFree cleaner, maybe?
      if (!IsConstant(gcd(*graeffeSeq, deriv(*graeffeSeq, IndetIndex))))
        return n;
    }
    return 0;
  }

  namespace // anonymous
  {
    /**
     * Used in modular approach to decide on whether
     * the n-th power Graeffe transform of f needs
     * to be checked for proving its LRS-degeneracy.
     * For private use. Hence, no arg checking.
     */
    bool ShouldCheckLRSExponent(ConstRefRingElem f, const unsigned long n)
    {
      VerboseLog VERBOSE("ShouldCheckLRSExponent");

      const ring &Px = owner(f);
      const ring &P = CoeffRing(Px);
      const long is = NumIndets(Px);
      PrimeSeq1ModN primeSeq(n);

      const unsigned long gcdBound = n / 2;

      // Magic number; try at most 3 primes
      for (int i = 0; i < 3; ++i)
      {
        const SmallPrime p = *primeSeq;
        if (IsZero(p)) // upper prime bound has been reached
          break;

        // zeta will be a primitive root of unity in F_p
        const QuotientRing Fp = NewZZmod(p);
        const SparsePolyRing Fpx = NewPolyRing(Fp, NewSymbols(1));
        const RingElem& x = indet(Fpx,0);
        const std::vector<RingElem> images(is, x);
        const RingElem f_p = PolyRingHom(Px, Fpx, CanonicalHom(P, Fp), images)(f);
        const unsigned long zeta = PowerMod(PrimitiveRoot(p), (p - 1) / n, p);
        VERBOSE(120) << "zeta: " << zeta << std::endl;

        for (unsigned long k = 1; k <= gcdBound; ++k)
        {
          CheckForInterrupt("ShouldCheckLRSExponent");
          if (IsCoprime(n, k) &&
              (IsCoprime(f_p, PolyAlgebraHom(Fpx, Fpx, std::vector<RingElem>({PowerMod(zeta, k, p) * x}))(f_p)) ||
               IsCoprime(f_p, PolyAlgebraHom(Fpx, Fpx, std::vector<RingElem>({PowerMod(zeta, n - k, p) * x}))(f_p))))
          {
            return false;
          }
        }
        ++primeSeq;
      }

      return true;
    }

    bool IsFactorOfOne(const unsigned long n, const std::set<unsigned long> &L)
    {
      for (unsigned long m : L)
      {
        if (m >= n && n % m == 0)
          return true;
      }
      return false;
    }

    /**
     * Used in modular approach to decide on whether
     * the n-th power Graeffe transform of f needs
     * to be checked for proving its LRS-degeneracy.
     * For private use.  Hence, no arg checking.
     */

    std::vector<unsigned long> GetKToTryList(const unsigned long D, const unsigned long bound)
    {
      // Create list of values which phi(k) must divide.
      // List is repr as a "boolean mask": mask[k] == 1 iff k is in list
      std::vector<unsigned char> mask(D*D);
      for (unsigned long d=2; d <= D; ++d)
        mask[d*(d-1)] = 1;
            
      for (unsigned long a = 1; a < D; ++a)
      {
        for (unsigned long b = 1; b <= a; ++b)
        {
          if (a+b <= D && IsEven(a * b))
            mask[a * b] = 1;
        }
      }

      // Find minimal subset of list above
      // Note loop considers values in descending order.
      std::set<unsigned long> NewL;
      for (unsigned long j = D*(D-1); j > 1; --j)
      {
        if (mask[j] == 0) continue;
        if (!IsFactorOfOne(j, NewL))
          NewL.insert(j);
      }


      // Find all k values such that phi(k) divides some elem of NewL
      std::vector<unsigned long> KToTry;
      for (unsigned long n = 2; n <= bound; ++n)
      {
        if (IsFactorOfOne(EulerTotient(n), NewL))
          KToTry.push_back(n);
      }
      return KToTry;
    }

  }

  /**
   * Checks for LRS-degeneracy of the given order, without arg checks
   */
  bool IsLRSDegenerateOrder(ConstRefRingElem f, const unsigned long n, const long IndetIndex)
  {
    VerboseLog VERBOSE("IsLRSDegenerateOrder");
    if (n < 2)
      CoCoA_THROW_ERROR(ERR::OutOfRange, "IsLRSDegenerateOrder");

    if (n == 2)
    {
      const ring &Px = owner(f);
      const RingElem &x = indet(Px, IndetIndex);
      // Equal to !IsCoprime
      if (!IsConstant(gcd(f, PolyAlgebraHom(Px, Px, std::vector<RingElem>(NumIndets(Px), -x))(f))))
        return true;
    }

    if (ShouldCheckLRSExponent(f, n))
    {
      VERBOSE(20) << "Exponent " << n << " could not be excluded" << std::endl;

      // Also possible: Res_y(f(y), cyclotomic(n,y))
      const RingElem Pn = GraeffeN(f, n); // equal to resultant(eval(f, [y]), y^n-x, y);
      if (!IsSqFree(Pn))
        return true;
    }

    return false;
  }

  /**
   * Checks for LRS-degeneracy of the given order
   */
  bool IsLRSDegenerateOrder(RingElem f, const unsigned long n)
  {
    VerboseLog VERBOSE("IsLRSDegenerateOrder");
    const ring &Px = owner(f);
    if (!IsSparsePolyRing(Px))
      CoCoA_THROW_ERROR(ERR::NotSparsePolyRing, "IsLRSDegenerateOrder");
    const ring &P = CoeffRing(Px);
    const long IndetIndex = UnivariateIndetIndex(f);
    if (IndetIndex < 0)
      CoCoA_THROW_ERROR(ERR::NotUnivariate, "IsLRSDegenerateOrder");

    if (IsQQ(P))
    {
      const ring &ZZ = RingZZ();
      const SparsePolyRing ZZx = NewPolyRing(ZZ, NewSymbols(NumIndets(Px)));
      const RingHom toZZx = PolyRingHom(Px, ZZx, QQEmbeddingHom(ZZ), indets(ZZx));
      return IsLRSDegenerateOrder(toZZx(f / content(f)), n);
    }

    if (!IsZZ(P))
      CoCoA_THROW_ERROR(ERR::BadRing, "IsLRSDegenerateOrder");

    const RingElem &x = indet(Px, IndetIndex);

    // Must have non-zero constant coefficient
    f /= gcd(f, power(x, deg(f)));

    const long degf = deg(f);
    if (degf < 2)
      CoCoA_THROW_ERROR("Polynomial does not have two distinct non-zero roots.", "IsLRSDegenerateOrder");

    if (!IsSqFree(f))
      CoCoA_THROW_ERROR("Polynomial is not square-free!", "IsLRSDegenerateOrder");

    return IsLRSDegenerateOrder(f, n, IndetIndex);
  }

  /**
   * Modular approach
   */
  unsigned long IsLRSDegenerateMod(RingElem f)
  {
    VerboseLog VERBOSE("IsLRSDegenerateMod");
    const ring &Px = owner(f);
    if (!IsSparsePolyRing(Px))
      CoCoA_THROW_ERROR(ERR::NotSparsePolyRing, "IsLRSDegenerateMod");
    const ring &P = CoeffRing(Px);
    const long IndetIndex = UnivariateIndetIndex(f);
    if (IndetIndex < 0)
      CoCoA_THROW_ERROR(ERR::NotUnivariate, "IsLRSDegenerateMod");

    if (IsQQ(P))
    {
      const ring &ZZ = RingZZ();
      const SparsePolyRing ZZx = NewPolyRing(ZZ, NewSymbols(NumIndets(Px)));
      const RingHom toZZx = PolyRingHom(Px, ZZx, QQEmbeddingHom(ZZ), indets(ZZx));
      return IsLRSDegenerateMod(toZZx(f / content(f)));
    }

    if (!IsZZ(P))
      CoCoA_THROW_ERROR(ERR::BadRing, "IsLRSDegenerateMod");

    const RingElem &x = indet(Px, IndetIndex);

    // TODO: Is there a way to get least non-zero degree?
    // Must have non-zero constant coefficient
    f /= gcd(f, power(x, deg(f)));

    const long degf = deg(f);
    if (degf < 2)
      CoCoA_THROW_ERROR("Polynomial does not have two distinct non-zero roots.", "IsLRSDegenerateMod");

    if (!IsSqFree(f))
      CoCoA_THROW_ERROR("Polynomial is not square-free!", "IsLRSDegenerateMod");

    if (IsLRSDegenerateOrder(f, 2, IndetIndex))
      return 2;

    BigInt bound = InversePhiBound(degf); // * degf ??? TODO
    unsigned long N;
    if (!IsConvertible(N, bound))
      CoCoA_THROW_ERROR(ERR::ArgTooBig, "IsLRSDegenerateMod");

    VERBOSE(20) << "n bound = " << N << std::endl;

    std::vector<unsigned long> KToTry = GetKToTryList(degf, N);
    for (unsigned long n : KToTry)
    {
      VERBOSE(100) << "n = " << n << std::endl;
      if (IsLRSDegenerateOrder(f, n, IndetIndex))
        return n;
    }
    return 0;
  }

  /**
   * Chooses the right approach for the given polynomial
   */
  unsigned long IsLRSDegenerate(ConstRefRingElem f)
  {
    return IsLRSDegenerateMod(f);
  }

}


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SparsePolyOps-LRSDegeneracy.C,v 1.5 2022/11/30 20:38:08 abbott Exp $
// $Log: SparsePolyOps-LRSDegeneracy.C,v $
// Revision 1.5  2022/11/30 20:38:08  abbott
// Summary: Fixed subtle bug
//
// Revision 1.4  2022/11/30 15:19:12  abbott
// Summary: Many changes; revised indentation
//
// Revision 1.3  2022/11/24 21:11:29  abbott
// Summary: Added copyright notice
//
//
